import inspect
import threading

def is_generator_function(obj):
    """Return true if the object is a user-defined generator function."""
    CO_GENERATOR = 0x20
    return bool(((inspect.isfunction(obj) or inspect.ismethod(obj)) and
                 obj.func_code.co_flags & CO_GENERATOR))


def import_class(name):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


class After(object):
    """Causes all contained Pipelines to run after the given ones complete.

    Must be used in a 'with' block.
    """

    _local = threading.local()

    def __init__(self, *futures):
        """Initializer.

        Args:
          *futures: PipelineFutures that all subsequent pipelines should follow.
            May be empty, in which case this statement does nothing.
        """
        for f in futures:
            if not isinstance(f, Future):
                raise TypeError(
                    'May only pass PipelineFuture instances to After(). %r', type(f)
                )
        self._futures = set(futures)

    def __enter__(self):
        """When entering a 'with' block."""
        After._thread_init()
        After._local._after_all_futures.extend(self._futures)

    def __exit__(self, type, value, trace):
        """When exiting a 'with' block."""
        for future in self._futures:
            After._local._after_all_futures.remove(future)
        return False

    @classmethod
    def _thread_init(cls):
        """Ensure thread local is initialized."""
        if not hasattr(cls._local, '_after_all_futures'):
            cls._local._after_all_futures = []


class InOrder(object):
    """Causes all contained Pipelines to run in order.

    Must be used in a 'with' block.
    """

    _local = threading.local()

    @classmethod
    def _add_future(cls, future):
        """Adds a future to the list of in-order futures thus far.

        Args:
          future: The future to add to the list.
        """
        if cls._local._activated:
            cls._local._in_order_futures.add(future)

    def __init__(self):
        """Initializer."""

    def __enter__(self):
        """When entering a 'with' block."""
        # Reentrancy checking gives false errors in test mode since everything is
        # on the same thread, and all pipelines are executed in order in test mode
        # anyway, so disable InOrder for tests.
        InOrder._thread_init()
        assert not InOrder._local._activated, 'Already in an InOrder "with" block.'
        InOrder._local._activated = True
        InOrder._local._in_order_futures.clear()

    def __exit__(self, type, value, trace):
        """When exiting a 'with' block."""
        InOrder._local._activated = False
        InOrder._local._in_order_futures.clear()
        return False

    @classmethod
    def _thread_init(cls):
        """Ensure thread local is initialized."""
        if not hasattr(cls._local, '_in_order_futures'):
            cls._local._in_order_futures = set()
            cls._local._activated = False




from .models import Pipeline, Slot, Barrier
from django_q.tasks import async
from datetime import datetime
import util
from util import After, InOrder


class Future(object):

    def __init__(self, pipe):
        self._after_all_pipelines = {}
        self.output = pipe.output

    @property
    def is_done(self):
        return self._value.status == Slot.STATUS.FILLED

    @property
    def value(self):
        if self.is_done:
            return self._value


class Pipe(object):
    def __init__(self, *args, **kwargs):
        self.pk = None
        self.args = args
        self.kwargs = kwargs
        self.class_path = "%s.%s" % (self.__module__, self.__class__.__name__)
        self.output = None

    @classmethod
    def from_id(cls, id):
        pipeline = Pipeline.objects.get(id=id)
        klass = util.import_class(pipeline.class_path)
        obj = klass(*pipeline.args, **pipeline.kwargs)
        obj.pk = id
        obj.output = pipeline.output
        return obj

    def save(self, **kwargs):
        if self.pk:
            p, _ = Pipeline.objects.update_or_create(id=self.pk)
        else:
            p = Pipeline()

        p.class_path = self.class_path
        p.args = self.args
        p.kwargs = self.kwargs
        p.output = self.output

        for k in kwargs:
            setattr(p, k, kwargs[k])

        p.save()

        if not p.output:
            p.output = Slot.objects.create(filler=p)
            p.save()

        self.pk = p.id
        self.output = p.output

    def run(self):
        raise NotImplementedError()

    def start(self):
        self.save()
        async(evaluate, self.pk)


def notify_barrier(barrier):
    if all(slot.status == Slot.STATUS.FILLED for slot in barrier.blocking_slots.all()):
        barrier.status = Barrier.STATUS.FIRED
        barrier.triggered = datetime.utcnow()
        barrier.save()

        async(evaluate, barrier.target.pk)


def notify_barriers(slot):
    for barrier in slot.barrier_set.all():
        notify_barrier(barrier)


def fill_slot(filler, slot, value):
    slot.filler_id = filler.pk
    slot.value = value
    slot.status = Slot.STATUS.FILLED
    slot.save()
    notify_barriers(slot)
    filler.save(status=Pipeline.STATUS.DONE)

def evaluate(pipeline_pk):
    After._thread_init()
    InOrder._thread_init()
    InOrder._local._activated = False

    pipeline = Pipe.from_id(pipeline_pk)

    # FIXME: handle not generator case
    pipeline_is_generator = util.is_generator_function(pipeline.run)

    args = pipeline.args
    kwargs = pipeline.kwargs
    assert not any(isinstance(k, Slot) for k in args)
    assert not any(isinstance(kwargs[k], Slot) for k in kwargs)

    if not pipeline_is_generator:
        result = pipeline.run(*args, **kwargs)
        fill_slot(pipeline, pipeline.output, result)
        return

    last_sub_stage = None
    sub_stage_dict = {}
    sub_stage_ordering = []
    next_value = None

    pipeline_iter = pipeline.run(*args, **kwargs)

    while True:
        try:
            yielded = pipeline_iter.send(next_value)
        except StopIteration:
            break
        except Exception, e:
            raise

        assert isinstance(yielded, Pipe)
        assert yielded not in sub_stage_dict

        yielded.save()
        last_sub_stage = yielded
        next_value = Future(yielded)
        next_value._after_all_pipelines.update(
            After._local._after_all_futures
        )
        next_value._after_all_pipelines.update(
            InOrder._local._in_order_futures
        )
        sub_stage_dict[yielded] = next_value
        sub_stage_ordering.append(yielded)
        InOrder._add_future(next_value)

    if last_sub_stage:
        pipeline.output = last_sub_stage.output
        pipeline.save()
    else:
        fill_slot(pipeline, pipeline.output, None)
        return

    for sub_stage in sub_stage_ordering:
        future = sub_stage_dict[sub_stage]
        sub_stage.save()
        child_pipeline = sub_stage

        dependent_slots = set()
        for arg in child_pipeline.args:
            if isinstance(arg, Slot):
                dependent_slots.add(arg)
        for key, arg in child_pipeline.kwargs.iteritems():
            if isinstance(arg, Slot):
                dependent_slots.add(arg)

        for other_future in future._after_all_pipelines:
            slot = other_future.output
            dependent_slots.add(slot)

        # dependent_slots.add(future.output)

        barrier = Barrier.objects.create(
            target_id=child_pipeline.pk,
        )
        barrier.blocking_slots.add(*dependent_slots)
        notify_barrier(barrier)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-05-29 10:26
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_p', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='barrier',
            name='target',
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE, to='django_p.Pipeline'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-05-29 09:41
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Barrier',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('triggered', models.DateTimeField()),
                ('status', models.IntegerField(choices=[
                 (1, b'FIRED'), (0, b'WAITING')], default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Pipeline',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('class_path', models.CharField(max_length=255)),
                ('started', models.DateTimeField(auto_now_add=True)),
                ('finalized', models.DateTimeField(auto_now=True)),
                ('params', jsonfield.fields.JSONField(default={})),
                ('status', models.IntegerField(choices=[
                 (3, b'WAITING'), (2, b'RUN'), (1, b'DONE'), (0, b'ABORTED')], default=3)),
                ('current_attempt', models.IntegerField(default=0)),
                ('max_attempts', models.IntegerField(default=1)),
                ('next_retry_time', models.DateTimeField()),
                ('retry_message', models.TextField()),
                ('abort_message', models.TextField()),
                ('abort_requested', models.BooleanField(default=False)),
                ('parent_pipeline', models.ForeignKey(blank=True, null=True,
                                                      on_delete=django.db.models.deletion.CASCADE, related_name='children', to='django_p.Pipeline')),
                ('root_pipeline', models.ForeignKey(blank=True, null=True,
                                                    on_delete=django.db.models.deletion.CASCADE, related_name='descendants', to='django_p.Pipeline')),
            ],
        ),
        migrations.CreateModel(
            name='Slot',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('value', jsonfield.fields.JSONField(default={})),
                ('status', models.IntegerField(choices=[
                 (1, b'FILLED'), (0, b'WAITING')], default=0)),
                ('filled', models.DateTimeField(auto_now=True)),
                ('filler', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='django_p.Pipeline')),
            ],
        ),
        migrations.AddField(
            model_name='barrier',
            name='blocking_slots',
            field=models.ManyToManyField(to='django_p.Slot'),
        ),
        migrations.AddField(
            model_name='barrier',
            name='target',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to='django_p.Pipeline'),
        ),
        migrations.AlterUniqueTogether(
            name='slot',
            unique_together=set([('filler', 'name')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-06-02 09:08
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_p', '0004_auto_20160529_1059'),
    ]

    operations = [
        migrations.AddField(
            model_name='pipeline',
            name='output',
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.CASCADE, to='django_p.Slot'),
        ),
        migrations.AlterUniqueTogether(
            name='slot',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='slot',
            name='name',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-05-29 10:27
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_p', '0002_auto_20160529_1026'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='pipeline',
            name='current_attempt',
        ),
        migrations.RemoveField(
            model_name='pipeline',
            name='max_attempts',
        ),
        migrations.RemoveField(
            model_name='pipeline',
            name='next_retry_time',
        ),
        migrations.RemoveField(
            model_name='pipeline',
            name='retry_message',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-06-02 14:13
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('django_p', '0005_auto_20160602_0908'),
    ]

    operations = [
        migrations.AlterField(
            model_name='slot',
            name='filler',
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.CASCADE, to='django_p.Pipeline'),
        ),
        migrations.AlterField(
            model_name='slot',
            name='value',
            field=jsonfield.fields.JSONField(default=None),
        ),
    ]



from django.db import models
from jsonfield import JSONField
from model_utils import Choices
import util
# Create your models here.


class Pipeline(models.Model):
    STATUS = Choices(
        (3, 'WAITING', "WAITING"),
        (2, 'RUN', "RUN"),
        (1, 'DONE', "DONE"),
        (0, 'ABORTED', "ABORTED")
    )

    class_path = models.CharField(max_length=255)
    root_pipeline = models.ForeignKey(
        "Pipeline", null=True, blank=True, related_name="descendants")
    parent_pipeline = models.ForeignKey(
        "Pipeline", null=True, blank=True, related_name="children")

    started = models.DateTimeField(auto_now_add=True)
    finalized = models.DateTimeField(auto_now=True)
    output = models.ForeignKey("Slot", null=True)

    params = JSONField(default={})

    status = models.IntegerField(choices=STATUS, default=STATUS.WAITING)

    # current_attempt = models.IntegerField(default=0)
    # max_attempts = models.IntegerField(default=1)
    # next_retry_time = models.DateTimeField()
    # retry_message = models.TextField()

    abort_message = models.TextField()
    abort_requested = models.BooleanField(default=False)

    def _to_value(self, v):
        if v['type'] == "value":
            return v['value']
        elif v['type'] == "Slot":
            slot = Slot.objects.get(pk=v['slot_key'])
            if slot.status == Slot.STATUS.FILLED:
                return slot.value
            else:
                return slot

    def _from_value(self, v):
        from .pipeline import Future

        if isinstance(v, Future):
            v = v.output
        if isinstance(v, Slot):
            return {"type": "Slot", "slot_key": v.pk}
        else:
            return {"type": "value", "value": v}

    @property
    def args(self):
        return [self._to_value(k) for k in self.params['args']]

    @args.setter
    def args(self, value):
        self.params['args'] = [self._from_value(k) for k in value]

    @property
    def kwargs(self):
        return {k: self._to_value(self._kwargs[k]) for k in self.params['kwargs']}

    @kwargs.setter
    def kwargs(self, value):
        self.params['kwargs'] = {k: self._from_value(k) for k in value}

    @property
    def is_root_pipeline(self):
        return self.root_pipeline is None

    def __unicode__(self):
        return "%s:%s" % (self.class_path, self.pk)


class Slot(models.Model):
    STATUS = Choices(
        (1, 'FILLED', "FILLED"),
        (0, 'WAITING', "WAITING")
    )

    filler = models.ForeignKey(Pipeline, null=True)

    value = JSONField(default=None)
    status = models.IntegerField(choices=STATUS, default=STATUS.WAITING)

    filled = models.DateTimeField(auto_now=True)

    def __unicode__(self):
        return "%s:%s" % (self.filler, self.value if self.status == Slot.STATUS.FILLED else "waiting")


class Barrier(models.Model):
    PURPOSE = Choices(
        (2, "ABORT", "ABORT"),
        (1, "FINALIZE", "FINALIZE"),
        (0, "START", "START")
    )

    STATUS = Choices(
        (1, "FIRED", "FIRED"),
        (0, "WAITING", "WAITING")
    )

    target = models.OneToOneField(Pipeline)

    blocking_slots = models.ManyToManyField(Slot)
    triggered = models.DateTimeField(null=True, auto_now=True)
    status = models.IntegerField(choices=STATUS, default=STATUS.WAITING)

    def __unicode__(self):
        return unicode(self.target)

# class Status(models.Model):
#     root_pipeline = models.ForeignKey(Pipeline)
#     message = models.TextField()

#     updated = models.DateTimeField(auto_now=True)



from django.test import TestCase
from .models import Pipe

# Create your tests here.


class AddOne(Pipe):

    def run(self, number):
        return number + 1


class AddTwo(Pipe):

    def run(self, number):
        result = yield AddOne(number)
        yield AddOne(result)



from django.contrib import admin
from .models import Pipeline, Barrier, Slot

# Register your models here.


class PipelineAdmin(admin.ModelAdmin):
    list_display = ('id', 'class_path', 'root_pipeline',
                    'started', 'output', 'status')


class BarrierAdmin(admin.ModelAdmin):
    list_display = ('id', 'target', 'triggered', 'status')


class SlotAdmin(admin.ModelAdmin):
    list_display = ('id', 'filler', 'value', 'status', 'filled')

admin.site.register(Pipeline, PipelineAdmin)
admin.site.register(Barrier, BarrierAdmin)
admin.site.register(Slot, SlotAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-05-29 10:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_p', '0003_auto_20160529_1027'),
    ]

    operations = [
        migrations.AlterField(
            model_name='barrier',
            name='triggered',
            field=models.DateTimeField(null=True),
        ),
    ]
